# -*- coding: utf-8 -*-
"""Implementation of :class:`biblary.bibliography.storage.AbstractStorage` that stores on the local file system."""
import hashlib
import io
import pathlib
import typing as t

from ..entry import BibliographyEntry
from .abstract import AbstractStorage, FileType

__all__ = ('FileSystemStorage',)


class FileSystemStorage(AbstractStorage):
    """Implementation of :class:`biblary.bibliography.storage.AbstractStorage` that builds from a Bibtex file."""

    def __init__(self, filepath: pathlib.Path, *_, **__):
        """Construct a new instance.

        :param filepath: absolute filepath to the base folder where files will be stored.
        """
        self.filepath = pathlib.Path(filepath)

    @staticmethod
    def validate_file_type(file_type):
        """Validate the ``file_type``.

        :raises ``TypeError``: if the given ``file_type`` is not a valid ``FileType``.
        """
        if not isinstance(file_type, FileType):
            raise TypeError(f'file_type `{file_type}` is not a valid `FileType`.')

    def get_filepath(self, entry: BibliographyEntry, file_type: FileType) -> pathlib.Path:
        """Return the byte content of a file with the given type for the given bibliographic entry.

        :param entry: the :class:`biblary.bibliographic.entry.BibliographicEntry` for which to retrieve the file.
        :param file_type: the file type to retrieve for the given entry.
        :returns: the absolute filepath where the file should be stored if it exists.
        """
        self.validate_file_type(file_type)

        doi_hash = hashlib.sha256(str(entry.identifier).encode('utf-8')).hexdigest()

        if isinstance(file_type, FileType):
            filename = file_type.value
        else:
            filename = file_type

        return self.filepath / doi_hash / filename

    def get_file(self, entry: BibliographyEntry, file_type: FileType) -> bytes:
        """Return the byte content of a file with the given type for the given bibliographic entry.

        :param entry: the :class:`biblary.bibliographic.entry.BibliographicEntry` for which to retrieve the file.
        :param file_type: the file type to retrieve for the given entry.
        :raises ``FileNotFoundError``: if the file of the given type does not exist for the given entry.
        :raises ``TypeError``: if the given ``file_type`` is not a valid ``FileType``.
        """
        with self.get_filepath(entry, file_type).open('rb') as handle:
            return handle.read()

    def put_file(self, content: t.Union[io.BytesIO, bytes], entry: BibliographyEntry, file_type: FileType) -> None:
        """Write the given byte content for the given bibliographic entry and file type.

        :param entry: the :class:`biblary.bibliographic.entry.BibliographicEntry` for which to retrieve the file.
        :param file_type: the file type to retrieve for the given entry.
        :raises ``TypeError``: if the ``content`` is not a byte-stream or pure bytes.
        """
        if isinstance(content, bytes):
            data = content
        else:
            data = content.read()

        if not isinstance(data, bytes):
            raise TypeError(f'invalid type for ``content``, should be bytes or byte-stream but got: `{content}`.')

        filepath = self.get_filepath(entry, file_type)
        filepath.parent.mkdir(parents=True, exist_ok=True)

        with filepath.open('wb') as handle:
            handle.write(data)
            handle.flush()

    def exists(self, entry: BibliographyEntry, file_type: FileType) -> bool:
        """Return whether the file with the given type for the given bibliographic entry exists.

        :param entry: the :class:`biblary.bibliographic.entry.BibliographicEntry` for which to retrieve the file.
        :param file_type: the file type to retrieve for the given entry.
        :returns: True if the file exists and False otherwise.
        :raises ``TypeError``: if the given ``file_type`` is not a valid ``FileType``.
        """
        return self.get_filepath(entry, file_type).exists()



# -*- coding: utf-8 -*-
"""Module that defines the URLs of this application."""
from django.urls import path

from .views import BiblaryBibtexView, BiblaryFileView, BiblaryIndexView, BiblaryUploadEntryView, BiblaryUploadFileView

app_name = 'biblary'  # pylint: disable=invalid-name

urlpatterns = [
    path('', BiblaryIndexView.as_view(), name='index'),
    path('upload-entry', BiblaryUploadEntryView.as_view(), name='upload-entry'),
    path('upload-file', BiblaryUploadFileView.as_view(), name='upload-file'),
    path('bibtex/<identifier>', BiblaryBibtexView.as_view(), name='bibtex'),
    path('file/<identifier>/<file_type>', BiblaryFileView.as_view(), name='file'),
]



# -*- coding: utf-8 -*-
"""Module with resources to model a bibliography."""
from .bibliography import Bibliography
from .entry import BibliographyEntry

__all__ = ('Bibliography', 'BibliographyEntry')



# -*- coding: utf-8 -*-
"""Abstract class that represents the interface to a file store of files associated with bibliographic entries."""
import abc
import enum
import io
import typing as t

from ..entry import BibliographyEntry

__all__ = ('AbstractStorage', 'FileType')


class FileType(enum.Enum):
    """Enumeration with the supported file types.

    Each bibliographic entry can have one file for each of these types.
    """

    MANUSCRIPT = 'manuscript'
    PREPRINT = 'preprint'
    SUPPLEMENTARY = 'supplementary'


class AbstractStorage(abc.ABC):
    """Abstract class that represents the interface to a file store of files associated with bibliographic entries."""

    @abc.abstractmethod
    def get_file(self, entry: BibliographyEntry, file_type: FileType) -> bytes:
        """Return the byte content of a file with the given type for the given bibliographic entry.

        :param entry: the :class:`biblary.bibliographic.entry.BibliographicEntry` for which to retrieve the file.
        :param file_type: the file type to retrieve for the given entry.
        :raises ``FileNotFoundError``: if the file of the given type does not exist for the given entry.
        """

    @abc.abstractmethod
    def put_file(self, content: t.Union[io.BytesIO, bytes], entry: BibliographyEntry, file_type: FileType) -> None:
        """Write the given byte content for the given bibliographic entry and file type.

        :param entry: the :class:`biblary.bibliographic.entry.BibliographicEntry` for which to retrieve the file.
        :param file_type: the file type to retrieve for the given entry.
        :raises ``TypeError``: if the ``content`` is not a byte-stream or pure bytes.
        """

    @abc.abstractmethod
    def exists(self, entry: BibliographyEntry, file_type: FileType) -> bool:
        """Return whether the file with the given type for the given bibliographic entry exists.

        :param entry: the :class:`biblary.bibliographic.entry.BibliographicEntry` for which to retrieve the file.
        :param file_type: the file type to retrieve for the given entry.
        :returns: True if the file exists and False otherwise.
        """



# -*- coding: utf-8 -*-
"""Module that defines the views of this application."""
from importlib import import_module
import typing as t

from django.core.exceptions import ImproperlyConfigured

from .bibliography import Bibliography


class BibliographyMixin:
    """Mixin to construct the :class:`biblary.bibliography.bibliography.Bibliography` from configured settings."""

    @staticmethod
    def construct_class(classifier: t.Optional[str] = None, kwargs: t.Dict = None) -> t.Optional[t.Any]:
        """Construct instance of the class specified by the given classifier using the provided keyword arguments.

        :param classifier: fully-qualified classifier of the class to construct.
        :param kwargs: keyword arguments that are passed to the constructor of the class.
        :returns: the class instance if successfully imported and constructed.
        :raises :class`django.core.exceptions.ImproperlyConfigured`: if the module or class of the classifier cannot be
            imported, or if the construction of the loaded class fails for the provided keyword arguments.
        """
        if classifier is None:
            return None

        module_name, _, class_name = classifier.rpartition('.')

        try:
            module = import_module(module_name)
        except (ImportError, ValueError) as exc:
            raise ImproperlyConfigured(f'module of `{classifier}` cannot be imported.') from exc

        try:
            cls = getattr(module, class_name)
        except AttributeError as exc:
            raise ImproperlyConfigured(f'class of `{classifier}` cannot be imported.') from exc

        try:
            instance = cls(**kwargs or {})
        except Exception as exc:
            raise ImproperlyConfigured(f'failed to construct `{classifier}` with keyword arguments: {kwargs}') from exc

        return instance

    @classmethod
    def get_bibliography(cls, storage_required=False) -> Bibliography:
        """Construct the bibliography with bibliographic entries from the configured settings.

        :param storage_required: boolean to indicate whether a configured storage is requird.
        :raises :class`django.core.exceptions.ImproperlyConfigured`: if bibliography cannot be properly instantiated.
        :raises :class`django.core.exceptions.ImproperlyConfigured`: if no storage is configured and the argument
            ``storage_required`` is set to ``True``.
        """
        from biblary.settings import settings

        try:
            adapter = cls.construct_class(settings.bibliography_adapter, settings.bibliography_adapter_configuration)
        except ImproperlyConfigured as exc:
            raise ImproperlyConfigured(f'failed to construct the configured bibliography adapter: {exc}') from exc

        if adapter is None:
            raise ImproperlyConfigured('no bibliography adapter has been configured.')

        try:
            storage = cls.construct_class(settings.bibliography_storage, settings.bibliography_storage_configuration)
        except ImproperlyConfigured as exc:
            raise ImproperlyConfigured(f'failed to construct the configured bibliography storage: {exc}') from exc

        if storage is None and storage_required:
            raise ImproperlyConfigured('file storage for this bibliography is required, but none has been configured.')

        return Bibliography(adapter, storage=storage)



# -*- coding: utf-8 -*-
"""Module that abstracts the interface to a storage of files associated with bibliographic entries.

For example, this can be used to store the PDF of an article associated with a certain bibliography entry.
"""
from .abstract import AbstractStorage, FileType
from .file_system import FileSystemStorage

__all__ = ('AbstractStorage', 'FileType', 'FileSystemStorage')



# -*- coding: utf-8 -*-
"""Implementation of :class:`biblary.bibliography.adapter.BibliographyAdapter` that builds from a Bibtex file."""
import dataclasses
import io
import pathlib
import shutil
import tempfile
import typing as t

from bibtexparser import customization, load
from bibtexparser.bibdatabase import BibDatabase
from bibtexparser.bparser import BibTexParser
from bibtexparser.bwriter import BibTexWriter

from ..entry import BibliographyEntry
from ..exceptions import BibliographicEntryParsingError
from .abstract import BibliographyAdapter


class BibtexBibliography(BibliographyAdapter):
    """Implementation of :class:`biblary.bibliography.adapter.BibliographyAdapter` that builds from a Bibtex file."""

    def __init__(self, filepath: pathlib.Path, *_, **__):
        """Construct a new instance.

        :param filepath: absolute filepath to a Bibtex file containing the bibliographic entries.
        """
        self.filepath = filepath

    @staticmethod
    def _transform_authors(record):
        """Reverse the ordering of the author parts and join with normal spaces.

        This will essentially transform "Oppenheimer, Robert" into "Robert Oppenheimer".
        """
        record['author'] = [' '.join(author.split(',')[::-1]).strip() for author in record['author']]
        return record

    @classmethod
    def _customize_record(cls, record):
        """Apply a set of transformations on the provided record."""
        transformers = (
            customization.convert_to_unicode,
            customization.keyword,
            customization.page_double_hyphen,
            customization.type,
            customization.author,
            cls._transform_authors,
        )

        for transformer in transformers:
            record = transformer(record)

        return record

    @staticmethod
    def _convert_entry(entry: t.Dict[str, t.Any]) -> BibliographyEntry:
        """Convert an entry parsed by ``bibtexparser`` into a ``BibliographyEntry``.

        :param entry: a dictionary representing the bibliographic entry.
        :return: the converted entry.
        """
        return BibliographyEntry(
            entry_type=entry['ENTRYTYPE'],
            identifier=entry['ID'],
            author=entry.get('author', None),
            title=entry.get('title', None),
            publisher=entry.get('publisher', None),
            journal=entry.get('journal', None),
            volume=entry.get('volume', None),
            issue=entry.get('issue', None),
            pages=entry.get('pages', None),
            month=entry.get('month', None),
            year=entry.get('year', None),
            keyword=entry.get('keyword', None),
            url=entry.get('url', None),
            doi=entry.get('doi', None),
        )

    @classmethod
    def _parse_bibliography(cls, filelike: t.TextIO) -> t.List[BibliographyEntry]:
        """Parse bibliographic entries from a text stream that should contain a ``.bib`` bibliography.

        :param filelike: a filelike object containing the content to parse.
        :return: list of parsed bibliographic entries.
        :raises :class:`bibliography.exceptions.BibliographicEntryParsingError`: if parsing fails.
        """
        parser = BibTexParser()
        parser.customization = cls._customize_record

        database = load(filelike, parser=parser)

        if not database.entries or not database.entries[0]:
            raise BibliographicEntryParsingError('failed to parse entries from bibliography.')

        return [cls._convert_entry(entry) for entry in database.entries]

    @classmethod
    def parse_entry(cls, content: str) -> BibliographyEntry:
        """Parse a new bibliographic entry from a string.

        :param content: the entry in string form.
        :return: the parsed bibliographic entry.
        :raises :class:`bibliography.exceptions.BibliographicEntryParsingError`: if parsing fails.
        """
        return cls._parse_bibliography(io.StringIO(content))[0]

    def get_entries(self) -> t.List[BibliographyEntry]:
        """Return the list of bibliography entries.

        :return: list of bibliographic entries.
        :raises :class:`bibliography.exceptions.BibliographicEntryParsingError`: if parsing fails.
        """
        with self.filepath.open() as handle:
            return self._parse_bibliography(handle)

    @staticmethod
    def _entry_to_dict(entry: BibliographyEntry) -> dict:
        """Convert a bibliographic entry to a dictionary.

        :param entry: bibliographic entry to write formatted to stream.
        :returns: entry in dictionary form.
        """
        dictionary = {
            'ENTRYTYPE': entry.entry_type,
            'ID': entry.identifier,
        }

        for field in dataclasses.fields(entry):
            if field.name in {'identifier', 'entry_type'}:
                continue

            value = getattr(entry, field.name)

            if value is not None:
                if field.name == 'author':
                    authors = ' and '.join([author.strip() for author in value])
                    dictionary[field.name] = authors
                else:
                    dictionary[field.name] = value

        return dictionary

    @classmethod
    def write_entry(cls, entry: BibliographyEntry, stream: t.TextIO) -> None:
        """Write an entry in bibtex format to the given stream.

        :param entry: bibliographic entry to write formatted to stream.
        """
        writer = BibTexWriter()
        writer.indent = '    '

        database = BibDatabase()
        database.entries.append(cls._entry_to_dict(entry))
        stream.write(writer.write(database))

    def save_entries(self, entries: t.List[BibliographyEntry]) -> None:
        """Save the list of entries to the bibliography.

        :param entries: list of bibliographic entries to write to the original bibliographic file.
        """
        database = BibDatabase()
        writer = BibTexWriter()
        writer.indent = '    '

        for entry in entries:
            database.entries.append(self._entry_to_dict(entry))

        with tempfile.NamedTemporaryFile('w') as handle:
            handle.write(writer.write(database))
            handle.flush()
            shutil.copy(handle.name, self.filepath)



# -*- coding: utf-8 -*-
"""Module with data class that represents an entry in a bibliography."""
from dataclasses import dataclass
import typing as t

__all__ = ('BibliographyEntry',)


@dataclass
class BibliographyEntry:
    """Class representing an entry in a bibliography."""

    entry_type: str
    identifier: str
    author: t.Optional[str] = None
    title: t.Optional[str] = None
    publisher: t.Optional[str] = None
    journal: t.Optional[str] = None
    volume: t.Optional[int] = None
    issue: t.Optional[int] = None
    pages: t.Optional[str] = None
    month: t.Optional[int] = None
    year: t.Optional[int] = None
    keyword: t.Optional[str] = None
    url: t.Optional[str] = None
    doi: t.Optional[str] = None



# -*- coding: utf-8 -*-
"""Abstract class representing the backend to a bibliography."""
import abc
import typing as t

from ..entry import BibliographyEntry


class BibliographyAdapter(abc.ABC):
    """Abstract class representing the backend to a bibliography."""

    @abc.abstractmethod
    def get_entries(self) -> t.List[BibliographyEntry]:
        """Return the list of bibliography entries."""

    @classmethod
    @abc.abstractmethod
    def parse_entry(cls, content: str) -> BibliographyEntry:
        """Parse a new bibliographic entry from a string.

        :param content: the entry in string form.
        :return: the parsed bibliographic entry.
        :raises :class:`bibliography.exceptions.BibliographicEntryParsingError`: if parsing fails.
        """

    @classmethod
    @abc.abstractmethod
    def write_entry(cls, entry: BibliographyEntry, stream: t.TextIO) -> None:
        """Write a bibliographic entry formatted as text to the given stream.

        :param entry: bibliographic entry to write formatted to stream.
        """

    @abc.abstractmethod
    def save_entries(self, entries: t.List[BibliographyEntry]) -> None:
        """Save the list of entries to the bibliography."""



# -*- coding: utf-8 -*-
"""Module with custom exceptions."""


class BibliographicEntryParsingError(ValueError):
    """Raised when :class:`bibliography.adapter.abstract.BibliographyAdapter.parse_entry` fails to parse the entry."""


class DuplicateEntryError(ValueError):
    """Raised when :class:`bibliography.bibliography.Bibliography.add_entry` receives duplicate entry."""


class InvalidBibliographyError(ValueError):
    """Raised when :class:`bibliography.bibliography.Bibliography` is constructed with an invalid bibliography."""



# -*- coding: utf-8 -*-
"""Module that defines the views of this application."""
import io
import typing as t

from django.core.exceptions import ImproperlyConfigured, SuspiciousOperation
from django.forms import Form
from django.http.response import Http404, HttpResponse
from django.urls import reverse_lazy
from django.views.generic import FormView, TemplateView, View

from .bibliography.adapter.bibtex import BibtexBibliography
from .bibliography.exceptions import BibliographicEntryParsingError, DuplicateEntryError
from .bibliography.storage import FileType
from .forms import BibliographyUploadEntryForm, BibliographyUploadFileForm
from .utils import BibliographyMixin


class BiblaryIndexView(BibliographyMixin, TemplateView):
    """View with index of bibliography contents."""

    template_name = 'biblary/index.html'

    def get_context_data(self, **kwargs):
        """Add the entries of the bibliography to the context."""
        bibliography = self.get_bibliography()

        context = super().get_context_data(**kwargs)
        context['entries'] = []

        for entry in bibliography.get_entries(sort=lambda entry: entry.year, reverse=True):

            if bibliography.storage is not None:
                entry.files = {file_type.value: bibliography.storage.exists(entry, file_type) for file_type in FileType}

            context['entries'].append(entry)

        return context


class BiblaryBibtexView(BibliographyMixin, View):
    """View that serves the bibliographic entry in bibtex format."""

    def get(self, _, *__, **___) -> HttpResponse:
        """Return the byte content of the bibliographic entry in bibtex format.

        :returns :class:`django.http.response.HttpResponse`: if the file exists for the specified entry and file type.
        :raises :class:`django.core.exceptions.SuspiciousOperation`: if the requested file type does not exist.
        :raises :class:`django.core.exceptions.Http404`: if the bibliographic entry does not exist, or it
            does but the requested file does not exist.
        """
        entry_identifier = self.kwargs['identifier']

        try:
            bibliography = self.get_bibliography(storage_required=True)
        except ImproperlyConfigured as exc:
            raise Http404('No files are available for the current configuration.') from exc

        assert bibliography.storage is not None

        try:
            entry = bibliography[entry_identifier]
        except KeyError as exc:
            raise Http404(f'The requested bibliographic entry `{entry_identifier}` does not exist.') from exc

        stream = io.StringIO()
        BibtexBibliography.write_entry(entry, stream)
        stream.seek(0)
        content = stream.read()

        return HttpResponse(
            content,
            headers={
                'Content-Type': 'application/plain',
                'Content-Disposition': f'attachment; filename="{entry_identifier}.bib"',
            }
        )


class BiblaryFileView(BibliographyMixin, View):
    """View that serves a file stored for a bibliographic entry."""

    def get(self, _, *__, **___) -> HttpResponse:
        """Return the byte content of the file for the specified bibliographic entry and file type.

        :returns :class:`django.http.response.HttpResponse`: if the file exists for the specified entry and file type.
        :raises :class:`django.core.exceptions.SuspiciousOperation`: if the requested file type does not exist.
        :raises :class:`django.core.exceptions.Http404`: if the bibliographic entry does not exist, or it
            does but the requested file does not exist.
        """
        entry_identifier = self.kwargs['identifier']
        file_type = self.kwargs['file_type']

        try:
            bibliography = self.get_bibliography(storage_required=True)
        except ImproperlyConfigured as exc:
            raise Http404('No files are available for the current configuration.') from exc

        assert bibliography.storage is not None

        try:
            file_type = FileType(self.kwargs['file_type'])
        except ValueError as exc:
            raise SuspiciousOperation(f'The requested file type `{file_type}` is invalid.') from exc

        try:
            entry = bibliography[entry_identifier]
        except KeyError as exc:
            raise Http404(f'The requested bibliographic entry `{entry_identifier}` does not exist.') from exc

        try:
            content = bibliography.storage.get_file(entry, file_type)
        except FileNotFoundError as exc:
            raise Http404(f'The requested file `{entry_identifier}:{file_type.value}` does not exist.') from exc

        return HttpResponse(
            content,
            headers={
                'Content-Type': 'application/pdf',
                'Content-Disposition': f'attachment; filename="{file_type.value}.pdf"',
            }
        )


class BiblaryUploadEntryView(BibliographyMixin, FormView):
    """View to upload a bibliographic entry."""

    form_class = BibliographyUploadEntryForm
    success_url = reverse_lazy('index')
    template_name = 'biblary/upload_entry.html'

    def get_form(self, form_class: Form = None) -> t.Optional[Form]:
        """Return an instance of the form to be used in this view if a file storage has been configured.

        Before returning the form, the choices of the ``entry_identifier`` are defined based on the configured and
        loaded bibliography. To upload files, a file storage is required. If no file storage is configured, ``None`` is
        returned instead of the form.
        """
        try:
            self.get_bibliography(storage_required=True)
        except ImproperlyConfigured:
            return None
        else:
            return super().get_form(form_class)

    def post(self, request, *args, **kwargs):
        """Validate that a bibliography with file storage has been configured and then forward the request.

        If no bibliography with file storage has been configured, forward to the ``get`` method which should display
        that the upload functionality is disables since no file storage is configured.
        """
        try:
            self.get_bibliography(storage_required=True)
        except ImproperlyConfigured:
            return super().get(request, *args, **kwargs)

        return super().post(request, *args, **kwargs)

    def form_valid(self, form: BibliographyUploadFileForm):
        """Attempt to add the content as an entry to the bibliography."""
        content = form.cleaned_data['content']
        bibliography = self.get_bibliography()

        try:
            bibliography.add_entry(content)
        except DuplicateEntryError as exception:
            form.add_error('content', exception)
            return super().form_invalid(form)
        except BibliographicEntryParsingError as exception:
            form.add_error('content', exception)
            return super().form_invalid(form)

        bibliography.save()

        return super().form_valid(form)


class BiblaryUploadFileView(BibliographyMixin, FormView):
    """View to upload a file of a give file type for a bibliographic entry."""

    form_class = BibliographyUploadFileForm
    success_url = reverse_lazy('upload-file')
    template_name = 'biblary/upload_file.html'

    def get_form(self, form_class: Form = None) -> t.Optional[Form]:
        """Return an instance of the form to be used in this view if a file storage has been configured.

        Before returning the form, the choices of the ``entry_identifier`` are defined based on the configured and
        loaded bibliography. To upload files, a file storage is required. If no file storage is configured, ``None`` is
        returned instead of the form.
        """
        try:
            bibliography = self.get_bibliography(storage_required=True)
        except ImproperlyConfigured:
            return None
        else:
            form = super().get_form(form_class)
            form.fields['entry_identifier'].choices = [(e.identifier, e.identifier) for e in bibliography.values()]
            return form

    def post(self, request, *args, **kwargs):
        """Validate that a bibliography with file storage has been configured and then forward the request.

        If no bibliography with file storage has been configured, forward to the ``get`` method which should display
        that the upload functionality is disables since no file storage is configured.
        """
        try:
            self.get_bibliography(storage_required=True)
        except ImproperlyConfigured:
            return super().get(request, *args, **kwargs)

        return super().post(request, *args, **kwargs)

    def form_valid(self, form: BibliographyUploadFileForm):
        """Write the uploaded content to the file storage configured for the loaded bibliography.

        .. note:: This method assumes that a bibliography including a file storage has been configured. This should have
            been checked in the ``post`` method of this view.
        """
        content = form.cleaned_data['content']
        file_type = form.cleaned_data['file_type']
        entry_identifer = form.cleaned_data['entry_identifier']

        bibliography = self.get_bibliography(storage_required=True)
        entry = bibliography[entry_identifer]
        assert bibliography.storage is not None

        with content.open('rb') as handle:
            bibliography.storage.put_file(handle, entry, file_type)

        return super().form_valid(form)



# -*- coding: utf-8 -*-
"""Module that defines a class through which application configuration settings can be retrieved."""
import typing as t


class Settings:
    """Container to provide configuration settings for the application."""

    def __init__(self, prefix: str):
        """Initialize the class.

        :param prefix: the prefix with which application settings can be configured globally. The prefix will be
            automatically converted to all uppercase.
        """
        self.prefix = prefix.upper()

    def _get_setting(self, name: str, default: t.Any) -> t.Any:
        """Retrieve the setting with the given name from the loaded settings or return the specified default.

        .. note:: The module :mod:`django.conf.settings` needs to be imported in this method for it to be up to date.
        """
        from django.conf import settings as django_settings
        return getattr(django_settings, f'{self.prefix}_{name}', default)

    @property
    def bibliography_adapter(self) -> str:
        """Return the full import path of the bibliography adapter implementation to use.

        The value should be an implementation of :class:`biblary.bibliography.adapter.abstract.BibliographyAdapter`.
        """
        return self._get_setting('BIBLIOGRAPHY_ADAPTER', 'biblary.bibliography.adapter.bibtex.BibtexBibliography')

    @property
    def bibliography_adapter_configuration(self) -> dict:
        """Return the dictionary that will be passed as keyword arguments of the bibliography adapter constructor."""
        return self._get_setting('BIBLIOGRAPHY_ADAPTER_CONFIGURATION', {})

    @property
    def bibliography_storage(self) -> t.Optional[str]:
        """Return the full import path of the bibliography storage implementation to use.

        The value should be an implementation of :class:`biblary.bibliography.storage.abstract.AbstractStorage`.
        """
        return self._get_setting('BIBLIOGRAPHY_STORAGE', None)

    @property
    def bibliography_storage_configuration(self) -> dict:
        """Return the dictionary that will be passed as keyword arguments of the bibliography storage constructor."""
        return self._get_setting('BIBLIOGRAPHY_STORAGE_CONFIGURATION', {})

    @property
    def bibliography_main_author_patterns(self) -> t.Sequence[str]:
        """Return a sequence of strings that represent authors that should be marked as main author.

        The elements of the sequence can be simple strings or regex patterns.
        """
        return self._get_setting('BIBLIOGRAPHY_MAIN_AUTHOR_PATTERNS', ())

    @property
    def bibliography_main_author_class(self) -> str:
        """Return the CSS class that is used by the :meth:`biblary.templatetags.authors.main_author_class`.

        This tag can be used in the index template to add this CSS class to main authors.
        """
        return self._get_setting('BIBLIOGRAPHY_MAIN_AUTHOR_CLASS', 'biblary-entry-author-main')


settings: Settings = Settings('BIBLARY')



# -*- coding: utf-8 -*-
"""Module with class that represents a bibliography."""
from collections.abc import Mapping
import typing as t

from .adapter import BibliographyAdapter
from .entry import BibliographyEntry
from .exceptions import DuplicateEntryError, InvalidBibliographyError
from .storage import AbstractStorage


class Bibliography(Mapping):
    """Collection of bibliographic entries.

    Entries are represented by instances of the :class:`biblary.bibliography.entry.BibliographyEntry` class. The parsing
    of the bibliographic entries is provided by an instance of :class:`biblary.bibliography.adapter.BibliographyAdapter`
    which is passed to the constructor.

    The class is implemented and behaves as a mapping. When iterated over it, it will return the identifiers of the
    bibliographic entries that it contains. The class can be indexed with an identifier to retrieve the corresponding
    entry from the collection.
    """

    def __init__(self, adapter: BibliographyAdapter, storage: t.Optional[AbstractStorage] = None):
        """Construct a new bibliography instance.

        :param adapter: bibliography adapter that provides access to the bibliography backend.
        """
        if not isinstance(adapter, BibliographyAdapter):
            raise TypeError(f'`adapter` should be an instance of `BibliographyAdapter`, but got: `{adapter}`.')

        self.adapter: BibliographyAdapter = adapter
        self.storage: t.Optional[AbstractStorage] = storage
        self._entries: t.Dict[str, BibliographyEntry] = self._initialize_entries()

    def __getitem__(self, key) -> BibliographyEntry:
        """Return a bibliographic entry for the given key which should correspond to the entry's identifier."""
        return self._entries[key]

    def __iter__(self) -> t.Iterator[str]:
        """Return an iterator over the bibliographic entries contained within this bibliography."""
        return iter(self._entries)

    def __len__(self) -> int:
        """Return the number of bibliographic entries contained within this bibliography."""
        return len(self._entries)

    def __contains__(self, entry: t.Any) -> bool:
        """Return whether the bibliography contains the given entry."""
        return entry.identifier in self._entries

    def refresh(self) -> None:
        """Refresh the state of the bibliography by parsing the current entries from the adapter.

        This should only have to be called if the bibliography source was updated through another ``Bibliography``
        instance. When the modification is done through the same instance, it is not necessary to refresh.
        """
        self._entries = self._initialize_entries()

    def _initialize_entries(self) -> t.Dict[str, BibliographyEntry]:
        """Initialize the internal mapping of bibliographic entries obtained through the adapter.

        :raises :class:`biblary.exceptions.InvalidBibliographyError`: if the bibliography contains entries with
            duplicate identifiers.
        """
        entries = self.adapter.get_entries()

        if entries is None:
            return {}

        identifiers = [entry.identifier for entry in entries]

        if len(identifiers) != len(set(identifiers)):
            raise InvalidBibliographyError('the configured bibliography contains entries with duplicate identifiers.')

        return {entry.identifier: entry for entry in entries}

    def get_entries(
        self,
        sort: t.Callable[[BibliographyEntry], int] = None,
        reverse: bool = False,
    ) -> t.List[BibliographyEntry]:
        """Return the list of bibliography entries.

        :param sort: optional lambda to sort the returned list of entries.
        :param reverse: whether to reverse the order of the sorting if `sort` is specified.
        """
        if sort is not None:
            return sorted(self._entries.values(), key=sort, reverse=reverse)

        return list(self._entries.values())

    def add_entry(self, entry: t.Union[BibliographyEntry, str]) -> BibliographyEntry:
        """Add a new entry.

        :param entry: the entry to add. If it is a ``str``, the method ``parse_entry`` of the adapter will be called to
            first parse the entry from the string content.
        :return: the entry that was added.
        :raises :class:`bibliography.exceptions.BibliographicEntryParsingError`: if parsing fails.
        :raises :class:`bibliography.exceptions.DuplicateEntryError`: if the bibliography already contains the entry.
        """
        if not isinstance(entry, BibliographyEntry):
            entry = self.adapter.parse_entry(entry)

        if entry in self:
            raise DuplicateEntryError(
                f'the bibliography already contains an entry with identifier `{entry.identifier}`.'
            )

        self._entries[entry.identifier] = entry

        return entry

    def save(self):
        """Persist the current state of the bibliography to the original source through the adapter."""
        self.adapter.save_entries(self.get_entries())



# -*- coding: utf-8 -*-
"""Module with template tags operating on the authors of bibliographic entries."""
import re
import typing as t

from django import template

register = template.Library()


@register.simple_tag()
def main_author_class(author: str) -> t.Optional[str]:
    """Return whether the given author matches any of the patterns in the ``BIBLIOGRAPHY_MAIN_AUTHOR_PATTERS``.

    :param author: the author.
    """
    from biblary.settings import settings

    patterns = settings.bibliography_main_author_patterns

    if not isinstance(patterns, tuple):
        raise TypeError('invalid configuration: setting `BIBLIOGRAPHY_MAIN_AUTHOR_PATTERS` should be a tuple.')

    if patterns and any(re.match(pattern, author) for pattern in patterns):
        return settings.bibliography_main_author_class

    return None



# -*- coding: utf-8 -*-
"""Module with form and form field definitions."""
from django import forms

from .bibliography.storage import FileType

__all__ = ('BibliographyUploadFileForm', 'BibliographyUploadEntryForm')


class FileTypeField(forms.TypedChoiceField):
    """Form field that provides a choice of a :class:`biblary.bibliography.storage.FileType`."""

    def __init__(self, *, choices=None, coerce=None, empty_value=None, **kwargs):
        """Construct a new instance.

        The ``choices``, ``coerce`` and ``empty_value`` argument of the ``forms.TypedChoiceField`` base class static and
        automatically determined from the :class:`biblary.bibliography.storage.FileType` enum. The constructor will
        raise a ``ValueError`` if they are specified.

        :raises ValueError: if ``choices``, ``coerce`` or ``empty_value`` are defined.
        """
        super().__init__(**kwargs)

        for arg in [choices, coerce, empty_value]:
            if arg is not None:
                raise ValueError(f'`{arg.__name__}` cannot be changed for the `FileTypeField`.')

        self.choices = [(file_type.value, file_type.value) for file_type in FileType]
        self.coerce = FileType
        self.empty_value = ()


class BibliographyUploadFileForm(forms.Form):
    """Form to upload a file of a given type for the specified bibligraphic entry.

    .. note:: The choices of the ``entry_identifier`` have to be specified in the view that presents the form because
        those should depend on the :class:`biblary.bibliography.Bibliography` that is configured and loaded. The values
        should correspond to the entries of the :class:`biblary.bibliography.entry.BibliographicEntry`'s.
    """

    entry_identifier = forms.ChoiceField(label='Entry')
    file_type = FileTypeField(label='Type')
    content = forms.FileField(label='File')


class BibliographyUploadEntryForm(forms.Form):
    """Form to upload a bibligraphic entry."""

    content = forms.CharField(label='', widget=forms.Textarea())
